/*****************************************************************************

  COPYRIGHT (c) 1995
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made
  without the express permission of the author. Any such copies, or
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author:
  Lydia Kavraki (kavraki@cs.stanford.edu)

******************************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/types.h>
#include <malloc.h>


/********************************************************************
	Define the robot specific functions used.
	The corresponding procedures sould be found 
	in the dependencies file. 
********************************************************************/

#define FIXED_LINKAGE	1

/**********for the fixed linkage*************************************/
#if FIXED_LINKAGE

#define VC(X, conf)             VC_fixed_linkage(X, conf)
#define auto_collision(conf, X) auto_collision_fixed_linkage(conf, X)    
#define FK(X, Y)                FK_fixed_linkage(X, Y) 
#define out_ws(X)               out_ws_fixed_linkage(X)
#define allign_robot(conf)      allign_robot_fixed_linkage(conf)
#define fill_in(X, conf, Xcl)   fill_in_fixed_linkage(X, conf, Xcl)
#define random_conf(X)		random_conf_fixed_linkage(X)
#define deviation_rand_conf(c_new, c_old, dev) \
	     deviation_rand_conf_fixed_linkage(c_new, c_old, dev)
#define connect_simple(conf1, conf2, flag_path, path_len, path) \
            connect_fixed_linkage(conf1, conf2, flag_path, path_len, path)
#define random_bounce(conff, max_length, length, path) \
            random_bounce_fixed_linkage(conff, max_length, length, path)
#define connect_line2(conf1, conf2, flag_path, path_len, path) \
	    connect_line2_fixed(conf1, conf2, flag_path, path_len, path)

#define draw_robot(conf, d, win, gc, scr, x_scale, y_scale) \
            draw_robot_fixed_linkage(conf, d, win, gc, scr, x_scale, y_scale)
#define undraw_robot(conf, d, win, gc, scr, x_scale, y_scale) \
            undraw_robot_fixed_linkage(conf, d, win, gc, scr, x_scale, y_scale)
#endif
     
     
/*******************************************************************
	Define the random, get time etc functions
********************************************************************/

#define mrandom(a)      (random() % (a))
	/*for the DECs. Measures CPU time. Problem: has a small period.*/
#define my_get_exact_time(t)  t = clock() 


/********************************************************************
	Definitions (a) for the workspace
		    (b) descritization
		    (c) graphics scale for debugging
*********************************************************************/

		/*discretization*/
#define DIM 		128
#define HALF_DIM	64
                /*used for the graphics windows*/
#define SCALE 		5
#define WIN_DIM         (DIM*SCALE) 
		/*for the environment*/
#define MAXOBST         20      /*max num of convex obstacles in workspace*/
#define MAXVERTEX       20      /*maximum number of vertices per obstacle */
#define MAX_LINK_VERTICES  10   /*maximun number of vertices each
                                  polygon link of the robot can have*/


/********************************************************************
        Definitions for the robot
*********************************************************************/


#if	FIXED_LINKAGE
	
#define DOF             7	/*always add 2 fro the base*/	
#define NBP		(DOF+1)
#define NUM_CNTL_POINTS 1       

#endif



/*******************************************************************
        The next lines should be changed when the maximun 
	number of intermediate poinst changes
*********************************************************************/

#define MAX_INTER 	 26000  /*max number of random confs*/	
#define MAX_COMP_INITIAL 10000	/*max no of comps after 1st connection*/
#define MAX_COMP	 10000 /*max no of comps after expansion*/		
	

/*******************************************************************
	The numbers below are the default values of several 
	parameters of the program.
	The parameters have the same nama. Their values are 
	changed in the programs.
*********************************************************************/

	/*up to this distance, connections are attempted.
	  Note that a good estimation of this value is of great
	  importance for the running time of the algorithm, that is 
	  why is should be determined experimentatly at a later stage.*/
#define BIG_DIST		20000	
	/*how many times the outside loop of connecting components 
	  is executed*/


#define MAXPATH         50000	/*max path length in number of 
				  configurations*/


/******************************************************************
	types of connections in the network
*******************************************************************/

#define	CONNECT_LINE		1	/*simple connection*/
#define CONNECT_RPP		2	/*RPP connection*/
#define CONNECT_RPP_LINE	3	/*random_bounce*/


/********************************************************************
	Do not change anything after this line 
********************************************************************/

typedef int configuration[DOF];


#define TWO_PI  	 6.283185
#define TICKS_PER_SECOND 1000000.0
#define MAXINT           1000000

		/*used when computing the skeleton and NF2 potential*/
#define DIMFRONT 	5000
#define DIMSKEL 	3000
#define FREE	        0 
#define OBSTACLE 	1
#define VISITED 	2
#define REVISITED 	3
#define IN_SKELETON 	4
#define IN_SKELETON_1	5 

		/*seeds*/
#define NUM_SEEDS 	1
#define MAX_NUM_SEEDS 	62
#define SEED_ARRAY  	{45,500,1000,2000,3000,4000,5000,6000,7000,8000,9000,\
		10000,11000,12000,13000,14000,15000,16000,17000,18000,\
		19000,20000,21000,22000,23000,24000,25000,26000,27000,\
		28000,29000,30000,31000,32000,33000,34000,35000,36000,\
		37000,38000,39000,40000,41000,42000,43000,44000,45000,\
		46000,47000,48000,49000,50000,51000,52000,53000,54000,\
		55000,56000,57000,58000,59000,60000}

		  /* used for approximations*/
#define	SMALL 1E-6
#define mask1 0x80000000  /* equal 2exp31 */
#define mask2 0xffffff80  /* equal 2exp31 - 127 */

/********************************************************************
	For debugging error messages	
*********************************************************************/

#define COMMENTS        1
#define ON_COMMENTS(x)  {if (COMMENTS) {x};}

enum datatype {INTEGER=1, REAL, STRING, CONFIGURATION, 
	BOOLEAN, NEW_LINE, END_GLOBALS};


/********************************************************************
	Data structure declarations
*********************************************************************/


typedef struct IPoint {
	int x,y;
}IPoint;

typedef struct IdPoint{
	int x,y;
	int id;
}IdPoint;

typedef struct FPoint{
    double x, y;
}FPoint;

typedef struct SklGraph{
  int x, y;
  int pointerToSuccessor;
} SklGraph;

typedef struct cell_path_k {
	int 	final, path_length;
	int	path_type;
	configuration *path;
	struct 	cell_path_k *next;
} cellp;


typedef struct data {
	char *prompt;
	void *addr;
	int type;
} data;



#include	"macro.h"
#include	"extern.h"
