/*****************************************************************************

  COPYRIGHT (c) 1995
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made
  without the express permission of the author. Any such copies, or
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: 
  Lydia Kavraki (kavraki@cs.stanford.edu)

******************************************************************************/



#include "defs.h"

#define 	DEBUG_PL 	0
#define 	CONNECT_CONF	0
#define 	CONNECT_CONF2	0
#define 	RANDOM_BOUNCE	0


#if  DEBUG_PL
        Display                 *display;
        int                     screen;
        Window                  win;
        GC                      gc;
        XGCValues 		gcvalues;
        char                    s[10];
#endif


/***********************************************************************
	compute the forward kinematics of the robot
***********************************************************************/

void FK_fixed_linkage(X, Y)
configuration X;
FPoint Y[NBP];
{
	int	i;

	Y[0].x = BaseX;
	Y[0].y = BaseY; /* first point OK */

	for(i=1; i<NBP; i++) {
	 Y[i].x = Y[i-1].x + Table_cos_length[ X[i-1] ];
	 Y[i].y = Y[i-1].y + Table_sin_length[ X[i-1] ];
	} /* other nodes of the linkage */
}



/************************************************************************
	Collision checking
	An approximation is done on the workspace posiition of
	the robot in order to use the Cspace bitmap.
**************************************************************************/
int VC_fixed_linkage(XX, conf)
FPoint XX[NBP];
configuration conf;
{
	int 		i;
	IPoint		X[NBP];


	double2int_ws_pos(XX, X);

	for(i=0; i<NBP-1; i++)
	 if(read_3d_bitmap(Cspace, X[i].x, X[i].y, conf[i]))
	  return 1;

		/*the following should be used when we do not 
		  want the robot to touch the obstacles - do 
		  not check the first link */
	for(i=0; i<NBP; i++) 
	 if(Dist[X[i].x][X[i].y] <= SAFETY)
	  return 1;

	return 0; 
}



/**************************************************************************
	Make sure the robot is not outside the workspace.
**************************************************************************/
int  out_ws_fixed_linkage(X)
FPoint X[NBP];
{
	int  i;
	
	for(i=0; i<NBP; i++) 
         if(X[i].x<0 || X[i].x>=DIM || X[i].y<0 || X[i].y>=DIM) {
	  return 1;
	 }
	return 0; 
}


/**********************************************************************
	The following adjusts all the angles of the robot
	in [0, DIM-1]
**********************************************************************/

void allign_robot_fixed_linkage(conf)
configuration conf;
{
        int i;
        for(i=0; i<DOF; i++)
         conf[i] = adjust_angle(conf[i]);
}


/**********************************************************************
	Autocollision (used macros in macro.h)	
**********************************************************************/

int     auto_collision_fixed_linkage(X, ws_pos)
configuration X;
FPoint  ws_pos[NBP];
{
        static int      angle_160;
        int             i, j, bool;


        ONCE_ONLY
         angle_160 = GRID(.5 * 160. / 180. * DIM);
        ONCE_END

	/* check that no two consecutive links are very close together */
	for(i=0; i<DOF-1; i++) {
	 check_in_cone(X[i], angle_160, X[i+1], bool);
	 if(!bool) return 1;
	}

	/* check that no two segments collide */
	for(i=2; i<NBP-1; i++)
	 for(j=0; j<i-1; j++) {
	  collide_seg(ws_pos[j], ws_pos[j+1], ws_pos[i], ws_pos[i+1],
			bool);
	  if(bool) return 1;
	 }

	/* this is because of discretization problems */
	point_near_segment(ws_pos[NBP-1], ws_pos[0], ws_pos[1], bool);
	if(bool) return 1;

        return 0;
}


/***************************************************************************
	creates a random configurations of the robot
***************************************************************************/
void random_conf_fixed_linkage(X)
configuration X;
{
	int	i;

	for(i=0; i<DOF; i++)
	 X[i] = mrandom(DIM);
}

void OLDdeviation_rand_conf_fixed_linkage(c_new, c_old, dev)
configuration	c_old, c_new;
int 		dev;
{
	int	j, ld; 

	ld = 2*dev + 1;
        for(j=0; j<DOF; j++)
         c_new[j] = adjust_angle(c_old[j] + (mrandom(ld) - dev));
}

void deviation_rand_conf_fixed_linkage(c_new, c_old, dev)
configuration	c_old, c_new;
int 		dev;
{
	int	j, ld; 

        for(j=0; j<DOF; j++) {
	 ld = (int)((j+1.)/DOF*dev);
         c_new[j] = adjust_angle(c_old[j] + (mrandom(1+2*ld) - ld));
	}
}



/*****************************************************************************
	Draws a position of the robot in the workspace with a specified 
	scale for the xaxis and the yaxis.
	Input is a configuration of the robot.
*****************************************************************************/
void draw_robot_fixed_linkage(conf, display, win, gc, screen, 
				xaxis_scale, yaxis_scale)
configuration conf;
Display* display;
Window win;
GC  gc;
int screen, xaxis_scale, yaxis_scale;
{
	int 	i;
	FPoint 	double_pos[NBP];
	IPoint  pos[NBP];
	XPoint	points[4];

	FK(conf, double_pos);
	double2int_ws_pos(double_pos, pos);
	XFlush(display);
	XSetForeground(display, gc, BlackPixel(display,screen));

	for(i=0; i<NBP-1; i++) {
    	 XDrawLine(display, win, gc, pos[i].x*xaxis_scale, 
				     pos[i].y*yaxis_scale, 
				     pos[i+1].x*xaxis_scale, 
			 	     pos[i+1].y*yaxis_scale);
  	 XFlush(display);
	}
        points[0].x = (pos[0].x)*xaxis_scale;
        points[0].y = (pos[0].y - 2)*yaxis_scale;

        points[1].x = (pos[0].x + 2)*xaxis_scale;
        points[1].y = (pos[0].y)*yaxis_scale;

        points[2].x = (pos[0].x)*xaxis_scale;
        points[2].y = (pos[0].y + 2)*yaxis_scale;

        points[3].x = (pos[0].x - 2)*xaxis_scale;
        points[3].y = (pos[0].y)*yaxis_scale;

        XFillPolygon(display, win, gc, points, 4,
                                Convex, CoordModeOrigin);

/*
	XFillRectangle(display, win, gc, pos[NBP-1].x*xaxis_scale,
                                     pos[NBP-1].y*yaxis_scale, 6, 6);
*/

  	XFlush(display);
}


/*****************************************************************************
        UnDraws a position of the robot in the workspace with a specified
        scale for the xaxis and the yaxis.
        Input is a configuration of the robot.
*****************************************************************************/
void undraw_robot_fixed_linkage(conf, display, win, gc, screen, 
				xaxis_scale, yaxis_scale)
configuration conf;
Display* display;
Window win;
GC  gc;
int screen, xaxis_scale, yaxis_scale;
{
        int     i;
        FPoint  double_pos[NBP];
	IPoint	pos[NBP];
	XPoint	points[4];

        FK(conf, double_pos);
	double2int_ws_pos(double_pos, pos);
        XFlush(display);
	XSetForeground(display, gc, WhitePixel(display,screen));
        for(i=0; i<NBP-1; i++) {
         XDrawLine(display, win, gc, pos[i].x*xaxis_scale,
                                     pos[i].y*yaxis_scale,
                                     pos[i+1].x*xaxis_scale,
                                     pos[i+1].y*yaxis_scale);
         XFlush(display);
        }
        points[0].x = (pos[0].x)*xaxis_scale;
        points[0].y = (pos[0].y - 2)*yaxis_scale;

        points[1].x = (pos[0].x + 2)*xaxis_scale;
        points[1].y = (pos[0].y)*yaxis_scale;

        points[2].x = (pos[0].x)*xaxis_scale;
        points[2].y = (pos[0].y + 2)*yaxis_scale;

        points[3].x = (pos[0].x - 2)*xaxis_scale;
        points[3].y = (pos[0].y)*yaxis_scale;

        XFillPolygon(display, win, gc, points, 4,
                                Convex, CoordModeOrigin);

/*
	XFillRectangle(display, win, gc, pos[0].x*xaxis_scale,
                                     pos[0].y*yaxis_scale, 6, 6);
*/
  	XFlush(display);
}


/****************************************************************************
	The following are used for implementing a simple
	but interesting planner for a robot which is 
	an articulated linkage with a moving base
****************************************************************************/
	
int	fill_in_fixed_linkage(X, conf, Xclose) /* returns 0 if it fails */
FPoint	X[NBP], Xclose[NBP];
configuration	conf;
{
	int	i, j, k, L = Link_length;
	double	x, y, l, nx, ny, theta;
	double	x1, y1, x2, y2, s, t, d1, d2;

	for(i=0; i<NBP-2; i+=2) {
	 /* find coordinates of i+1 */
	 x = .5*(X[i+2].x-X[i].x);
	 y = .5*(X[i+2].y-X[i].y);
	 l = x*x+y*y;
	 if(l<=0) return 0;
	 l = sqrt(l);
	 nx = -y/l;
	 ny = x/l;
	 l = L*L - l*l;
	 if(l<=0) return 0;
	 l = sqrt(l);
	 x1 = X[i].x + x + l*nx;
	 y1 = X[i].y + y + l*ny;
	 x2 = X[i].x + x - l*nx;
	 y2 = X[i].y + y - l*ny;
	 s = x1 - Xclose[i+1].x;
	 t = y1 - Xclose[i+1].y;
	 d1 = s*s+t*t;
	 s = x2 - Xclose[i+1].x;
	 t = y2 - Xclose[i+1].y;
	 d2 = s*s+t*t;
	 /* select the one that is close to the previous pos*/
	 if(d1<d2) {
	  X[i+1].x = x1;
	  X[i+1].y = y1;
	 }
	 else {
	  X[i+1].x = x2;
	  X[i+1].y = y2;
	 }
	 k = i;
	 theta = atan2(X[i+1].y-X[i].y, X[i+1].x-X[i].x);
	 conf[k] = theta2DIM(theta);
	 k++;
	 theta = atan2(X[i+2].y-X[i+1].y, X[i+2].x-X[i+1].x);
	 conf[k] = theta2DIM(theta);
	}
	if(NBP%2 == 0) {
	 /* last link moves freely */
	 /* assume the last angle is given in conf and fill in
	    only the workspace coords of the last node */
	 X[NBP-1].x = X[NBP-2].x + Table_cos_length[conf[DOF-1]];
	 X[NBP-1].y = X[NBP-2].y + Table_sin_length[conf[DOF-1]];
	}
	return 1;
}



int connect_fixed_linkage(conf1, conf2, flag_path, path_len, path)
configuration	conf1, conf2;
int		flag_path, *path_len;
configuration	path[MAXPATH];
{
	FPoint  	ws_pos1[NBP], ws_pos2[NBP], ws_pos[NBP];	
	FPoint 		previous_ws_pos[NBP]; 
	double		last_dof;
	FPoint  	step[NBP], double_ws_pos[NBP];
	double 		max_dist, points, tmp, length, step_angle;
	double		diff, abs_diff;
	configuration	conf, previous_conf;
	int 		i, j, flag, new_len, bool; 
	
	cmp_confs(conf1, conf2, bool);
	if(bool) {
         if(flag_path) {
         assign_conf(conf1, path[0]);
         *path_len = 1;
        }
        return 1;
	}

	if(flag_path) {
	 assign_conf(conf1, path[0]);
	 *path_len = 1;
	}
	FK(conf1, ws_pos1);
	FK(conf2, ws_pos2);

	/*determine the length of the longer line in the workspace*/
	max_dist = -1;
	for(i=0; i<NBP; i+=2) {
	 tmp  = SQUARE(ws_pos1[i].x - ws_pos2[i].x) +
	        SQUARE(ws_pos1[i].y - ws_pos2[i].y);
	 if(tmp > max_dist) max_dist = tmp;
	}

	/*determine the line along which the last dof moves
	  in cases it is needed, use the smallest angle*/
	length = 0;
	if(NBP%2 == 0) {
         diff = conf2[DOF-1]-conf1[DOF-1];
         abs_diff = my_ABS(diff);
         if(DIM-abs_diff > abs_diff) {
          length = abs_diff;
	  step_angle = diff;
	 }
         else {
          tmp = length = DIM-abs_diff;
          if(diff>0) step_angle = -tmp;
           else  step_angle = tmp;
         }
        }

	/*find the step increment along each line*/
	max_dist = sqrt(max_dist);
	points = GRID(my_MAX(max_dist, length) + 5);
	if(points < 1.0) {
	 printf("No need to connect.\n");
	 print_conf(conf1);
	 print_conf(conf2);
	 if(flag_path) *path_len = 0;
	 return 1;
	}
	for(i=0; i<NBP; i+=2) { 
	 step[i].x = (ws_pos2[i].x-ws_pos1[i].x)/points;
	 step[i].y = (ws_pos2[i].y-ws_pos1[i].y)/points;
	}
	if(NBP%2 == 0)  {
	 step_angle  /= points;
	}

	/*keep the previous configuration and ws_pos*/
	for(i=0; i<NBP; i++) {
	 double_ws_pos[i].x = previous_ws_pos[i].x = ws_pos1[i].x;	
	 double_ws_pos[i].y = previous_ws_pos[i].y = ws_pos1[i].y;	
	}
	if(NBP%2 == 0)  last_dof = conf1[DOF-1];
	for(i=0; i<DOF; i++) 
	 conf[i] = previous_conf[i] = conf1[i];

	/*advance the robot*/
	flag = 1;
        for(i=0; flag && i<points; i++) {
	 /*find the next position of some links in the workspace*/
	 for(j=0; j<NBP; j+=2) {
	   double_ws_pos[j].x += step[j].x;
	   double_ws_pos[j].y += step[j].y;
	 }
	 if(NBP%2 == 0) {
	  last_dof += step_angle;
	  if(last_dof>=DIM) last_dof -= DIM;
	  if(last_dof<=0) last_dof = DIM + last_dof; 
	  conf[DOF-1] = GRID(last_dof);
         }
 
		/*find the configuration*/
	 if(!fill_in_fixed_linkage(double_ws_pos, conf, previous_ws_pos)) {
	  flag = 0;
	  continue;
	 }
	 allign_robot(conf);
	 FK(conf, ws_pos);

		/*check the configuration*/
         if(out_ws(ws_pos) || (VC(ws_pos, conf)) ||
                auto_collision(conf, ws_pos)) {
          flag = 0;
         }
         else {
	  for(j=0; j<NBP; j++) {
           previous_ws_pos[j].x = double_ws_pos[j].x;
           previous_ws_pos[j].y = double_ws_pos[j].y;
          }
	  assign_conf(conf, previous_conf);

	  if(flag_path) {
	   assign_conf(conf, path[*path_len]);
	   (*path_len)++;
	  } 
	 } /*else*/

        } /*for i*/

	if(!flag) {
	 flag =  connect_line2(previous_conf, conf2, flag_path,
                                &new_len, &(path[*path_len][0]));
         if(flag_path) *path_len = *path_len + new_len;
        }
	else {
		/*check carefully*/
	 max_dist = -1;
         for(i=0; i<NBP; i++) {
          tmp = SQUARE(ws_pos[i].x - ws_pos2[i].x) +
                SQUARE(ws_pos[i].y - ws_pos2[i].y);
          if(tmp > max_dist) max_dist = tmp;
	 }

		/*depending on the planner determine failure */
	 if(max_dist>2) { 
		/*straight line not used*/
		/*straight line used */
	  flag =  connect_line2(previous_conf, conf2, flag_path,
                                &new_len, &(path[*path_len][0]));
#if     CONNECT_CONF2
	ONCE_ONLY
        display = XOpenDisplay(NULL);
        screen = DefaultScreen(display);
	gcvalues.line_width = 2;
	gcvalues.foreground = BlackPixel(display, screen);
        gc = XCreateGC(display, RootWindow(display,screen), 0, &gcvalues);
        win = XCreateSimpleWindow(display, RootWindow(display, screen),
                            0,0,640,640,DefaultDepth(display,screen),
                            BlackPixel(display, screen),
                            WhitePixel(display, screen));
        XSelectInput(display, win, ExposureMask | ButtonPressMask);
        XMapWindow(display,win);
        XFlush(display);
        printf("Press <return>");
        gets(s);
	ONCE_END
        XSetForeground(display,gc,WhitePixel(display,screen));
        XFillRectangle(display,win,gc,0,0,641,641);
        XFlush(display);
        XSetForeground(display,gc,BlackPixel(display,screen));
        draw_workspace(display, win, gc, screen, SCALE, SCALE,
                             ObjNum, ObjectwVertice, Objectw );
	draw_robot(conf1, display, win, gc, screen, SCALE, SCALE);
	draw_robot(conf2, display, win, gc, screen, SCALE, SCALE);
	gets(s);
	for(i=0; i<*path_len; i++) {
         draw_robot(path[i], display, win, gc, screen, SCALE, SCALE);
	}
        printf("Press <return>");
        gets(s);
	for(i=*path_len ; i<*path_len+new_len; i++) {
	draw_robot(path[i], display, win, gc, screen, SCALE, SCALE);
	}
	printf("Press <return>");
	gets(s);
#endif
          if(flag_path) *path_len = *path_len + new_len;
	} 
	} /*else*/

	if(flag_path) {
         assign_conf(conf2, path[*path_len]);
         (*path_len)++;
	}

        if(flag) return 1; else return 0;
}


/***************************************************************************/
void random_bounce_fixed_linkage(conff, max_length, length, path)
configuration 	conff;
int		*length, max_length;
configuration	path[MAXPATH];
{

	double	step[DOF], double_conf[DOF];	
	int	i, k, s, times, flag;
	FPoint	ws_pos[NBP];
	configuration	previous_conf, conf;
	char	ss[10];
	long	t1, t2;

#if     RANDOM_BOUNCE 
	ONCE_ONLY
        display = XOpenDisplay(NULL);
        screen = DefaultScreen(display);
	gcvalues.line_width = 2;
	gcvalues.foreground = BlackPixel(display, screen);
        gc = XCreateGC(display, RootWindow(display,screen), 0, &gcvalues);
        win = XCreateSimpleWindow(display, RootWindow(display, screen),
                            0,0,640,640,DefaultDepth(display,screen),
                            BlackPixel(display, screen),
                            WhitePixel(display, screen));
        XSelectInput(display, win, ExposureMask | ButtonPressMask);
        XMapWindow(display,win);
        XFlush(display);
        printf("Press <return>");
        gets(ss);
	ONCE_END
        XSetForeground(display,gc,WhitePixel(display,screen));
        XFillRectangle(display,win,gc,0,0,641,641);
        XFlush(display);
        XSetForeground(display,gc,BlackPixel(display,screen));
        draw_workspace(display, win, gc, screen, SCALE, SCALE,
                             ObjNum, ObjectwVertice, Objectw );
        draw_robot(conff, display, win, gc, screen, SCALE, SCALE);
        printf("Max length %d", max_length);
        printf("Press <init return>");
        gets(ss);
#endif

		/*pick a random direction*/
/*
	my_get_exact_time(t1);
*/
	assign_conf(conff, conf);
	assign_conf(conff, previous_conf);
	*length  = 0;
	times = 0;

	while((max_length>0) && (times<1000)) {

	for(i=0; i<DOF; i++) {
	 if(mrandom(2) == 0) step[i] = -1.0*(i+1)/DOF;
	 else step[i] = 1.0*(i+1)/DOF;
	}
		/*advance*/
	for(i=0; i<DOF; i++) double_conf[i] = previous_conf[i]; 

	for(s=0, flag = 1; flag && s<max_length; s++) {

	 for(i=0; i<DOF; i++) {
	  double_conf[i] += step[i]; 
	  conf[i] = GRID(double_conf[i]);
         }
	 allign_robot(conf);
	 FK(conf, ws_pos);

		/*check the configuration*/
         if(out_ws(ws_pos) || (VC(ws_pos, conf)) ||
                auto_collision(conf, ws_pos)) {	
          flag = 0;
         }
	 else {
	  assign_conf(conf, previous_conf); 
	  assign_conf(conf, path[*length]);
	  (*length)++;
	 }
	}

	max_length -= s;
	times++;
	}
/*
	my_get_exact_time(t2);
*/

#if     RANDOM_BOUNCE 
        XSetForeground(display,gc,WhitePixel(display,screen));
        XFillRectangle(display,win,gc,0,0,641,641);
        XFlush(display);
        XSetForeground(display,gc,BlackPixel(display,screen));
        draw_workspace(display, win, gc, screen, SCALE, SCALE,
                             ObjNum, ObjectwVertice, Objectw );
	for(i=0; i<*length; i++) {
         draw_robot(path[i], display, win, gc, screen, SCALE, SCALE);
	}
	if(*length ==0) printf("No path\n");
        printf("Press <return>");
        gets(ss);
#endif
}


		/*try to connect the angles that seem the closest*/
int connect_line2_fixed(conf1, conf2, flag_path, path_len, path)
configuration 	conf1, conf2;
int		flag_path, *path_len;
configuration	path[MAXPATH];
{
        int     i, j, flag, diff, abs_diff, points, tmp, ii;
        int     int_conf[DOF];
        double  double_conf[DOF], der[DOF], length;
        FPoint  ws_son[NBP];
	static	first_time = 1;

	if(flag_path) {
	 assign_conf(conf1, path[0]);
	 *path_len = 1;
	}

	length = 0;
/*
        for(i=0; i<3; i++) { 
	 tmp = conf2[i] - conf1[i];
         der[i] = tmp ;
	 length += SQUARE(tmp);
	}
*/
	
		/*do a heuristic on the angles*/
	for(i=0; i<DOF; i++) {
	 diff = conf2[i]-conf1[i];
	 abs_diff = my_ABS(diff);
	 if(DIM-abs_diff > abs_diff) { 
	  der[i] = diff;
	  length += SQUARE(diff);
	 }
	 else {
	  tmp = DIM-abs_diff;
	  length += SQUARE(tmp);
	  if(diff>0) der[i] = -tmp;
	   else  der[i] = tmp; 
	 }
	}
	points = GRID(1.5*sqrt(length));
	if(points == 0) {
         printf("No need to connect.\n");
         print_conf(conf1);
         print_conf(conf2);
	 if(flag_path) *path_len = 0;
         return 1;
        }

	for(i=0; i<DOF; i++) der[i] /= points;

                /*keep an int and a real conf*/
        assign_conf(conf1, int_conf);
        for(i=0; i<DOF; i++) double_conf[i] = conf1[i];

        flag = 1;
        for(i=0; i<points; i++) {
                /*check that this is a valid conf*/
         FK(int_conf, ws_son);
         if(out_ws(ws_son) || (VC(ws_son, int_conf)) || 
		auto_collision(int_conf, ws_son) ) { 
          flag = 0;
          break;
         }
#if     DEBUG_CONNECT2
        draw_robot(int_conf, display, win, gc, screen, SCALE, SCALE);
#endif
	 if(flag_path && (i%2 == 0)) {	
	  assign_conf(int_conf, path[*path_len]);
	  (*path_len)++;
	 }
	 
                       /*create the next configuration*/
         for(j=0; j<DOF; j++) {
          double_conf[j] += der[j];
          int_conf[j] = ROUND(double_conf[j]);
         }
	 allign_robot(int_conf);
        }

	if(flag_path) {
	 assign_conf(conf2, path[*path_len]);
	 (*path_len)++;
	}

#if     DEBUG_LINE2_ONLY
        printf("Path len = %d\n", *path_len);
        print_conf(conf1);
        print_conf(conf2);
        if(first_time) {
        display = XOpenDisplay(NULL);
        screen = DefaultScreen(display);
        gcvalues.line_width = 2;
        gcvalues.foreground = BlackPixel(display, screen);
        gc = XCreateGC(display, RootWindow(display,screen), 0, &gcvalues);
        win = XCreateSimpleWindow(display, RootWindow(display, screen),
                            0,0,640,640,DefaultDepth(display,screen),
                            BlackPixel(display, screen),
                            WhitePixel(display, screen));
        XSelectInput(display, win, ExposureMask | ButtonPressMask);
        XMapWindow(display,win);
        XFlush(display);
        first_time = 0;
        }
	if(flag) {
        printf("Press <return>");
        gets(s);
        XSetForeground(display,gc,WhitePixel(display,screen));
        XFillRectangle(display,win,gc,0,0,641,641);
        XFlush(display);
        XSetForeground(display,gc,BlackPixel(display,screen));
        draw_workspace(display, win, gc, screen, SCALE, SCALE,
                             ObjNum, ObjectwVertice, Objectw );
        draw_robot(conf1, display, win, gc, screen, SCALE, SCALE);
        draw_robot(conf2, display, win, gc, screen, SCALE, SCALE);
	
        for(ii=0; ii<*path_len; ii++)
        draw_robot(path[ii], display, win, gc, screen, SCALE, SCALE);
        printf("Press <return>");
        gets(s);
        free(path);
	}
#endif

        if(flag) return 1; else return 0;
}
