/*****************************************************************************

  COPYRIGHT (c) 1995 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Modified by Lydia Kavraki (kavraki@cs.stanford.edu)

******************************************************************************/

#include "defs.h"

typedef struct _config_node {
	configuration q;
	struct _config_node *next;
}config_node;

extern config_node *collision_free();
extern config_node *smooth2_aux();
extern config_node *append_config_node();
extern double sqr_euclidean_distance();
extern double curvilinear(); 
extern int shift_path();
extern int CurvePath();
extern int list_length();
extern void swap_int();
extern void free_config_node();



#define CONS(type,item,head) \
\
  item=(type *)calloc(1,sizeof(type)); \
  item->next=head; \
  head=item; \
\

#define assign(a,b) bcopy(a,b,sizeof(configuration))

#define assign_pos(a,b)\
  b.x=a.x;\
  b.y=a.y;


void INTERPOLATE(start_pt,end_pt,i,intervals,q) 
configuration start_pt, end_pt;
int i, intervals;
configuration q;
{ 
    int j_; 
    for (j_=0; j_<DOF; j_++) 
    q[j_]=GRID((((intervals-i)*start_pt[j_])+(i*end_pt[j_]))/(1.*intervals));
}

/****************************************************************************
  
  smooth4: multigrid resolution - sample progressively higher frequencies,
  systematically.
  
****************************************************************************/

void smooth4(configuration pX[MAXPATH], int *cur_path_len, int intervals ) 
{
  int path_len, prev_path_len, pt, radius, delta;
  int intervals2;
  config_node *head;
  int pt1, pt2;
  double factor;
  
  fprintf(stderr,"Smooth the path using method #4\n");
  factor=1.0;
  prev_path_len=0;
  path_len= *cur_path_len;
  /* outer loop */
  do {
    /* inner loop */
    delta=(int)(intervals/3);
    pt=delta;
    do {
      radius=(int)((factor*path_len)/2);
      pt1=my_MAX(0,pt-radius);
      pt2=my_MIN(path_len-1,pt+radius);
      intervals2=(int)(((pt2-pt1)/(2.0 * radius)) * intervals);
      prev_path_len=path_len;
      if ((head=collision_free(intervals2,pX,pt1,pt2,NULL))!=NULL) {
	path_len=shift_path(pX,path_len,head,pt1,pt2);
/*
	if (prev_path_len != path_len)
	  printf("prev_path_len=%d current path_len=%d\n", 
					prev_path_len, path_len);
*/
      }
      pt += delta;
    } while (pt < path_len);
    factor = factor/2.0;
    intervals = (int)(intervals/2);
  } while (delta > 2);
/*
  printf("Before CurvePath\n");
  CurvePath(pX, path_len);
*/
  *cur_path_len = path_len;
}

/***************************************************************************
  
  smooth3: multigrid resolution - sample progressively higher frequencies,
  randomly
  
****************************************************************************/

void smooth3(configuration pX[MAXPATH], int *cur_path_len, int intervals ) 
{
  int path_len, prev_path_len, random_pt, radius, same_count, count_limit;
  int intervals2;
  config_node *head;
  int pt1, pt2;
  double factor;
  
  printf("Smooth the path using method #3\n");
  factor=1.0;
  prev_path_len=0;
  path_len= *cur_path_len;
  count_limit=10;

  /* outer loop */
  do {
    same_count=0;
    /* inner loop */
    do {
      random_pt=random() % path_len;
      radius=(int)((factor*path_len)/2);
      pt1=my_MAX(0,random_pt-radius);
      pt2=my_MIN(path_len-1,random_pt+radius);
      intervals2=((pt2-pt1)/(2.0 * radius)) * intervals;
      prev_path_len=path_len;
      if ((head=collision_free(intervals2,pX,pt1,pt2,NULL))!=NULL) {
	path_len=shift_path(pX,path_len,head,pt1,pt2);
	printf("prev_path_len=%d path_len=%d\n", prev_path_len, path_len);
      };
      same_count=((prev_path_len==path_len) ? (same_count+1) : 0);
    } while (same_count < count_limit);
    factor /= 2.0;
    intervals /= 2;
    count_limit *= 2;
  } while (intervals > 2);
  *cur_path_len = path_len;
}


/***************************************************************************

  smooth2: random sampling of configuration space.

****************************************************************************/

void smooth2(configuration pX[MAXPATH], int *cur_path_len, int intervals ) 
{
  int path_len, prev_path_len;
  config_node *head;
  int point1, point2;

  prev_path_len=0;
  path_len= *cur_path_len;
  do {
    printf("prev_path_len=%d path_len=%d\n", prev_path_len, path_len);
    if ((head=smooth2_aux(intervals,pX,path_len,&point1,&point2))!=NULL) {
      prev_path_len=path_len;
      if (list_length(head)>(point2-point1)) {
	printf("interpolated list too long between %d and %d\n",
	       point1, point2);
	exit(-1);
      }
      path_len=shift_path(pX,path_len,head,point1,point2);
    };
  } while (prev_path_len != path_len);
  *cur_path_len = path_len;
}

config_node *smooth2_aux(intervals,pX,len,ppoint1,ppoint2)
     int intervals;
     configuration pX[MAXPATH];
     int len;
     int *ppoint1, *ppoint2;
{
  int point1, point2, path_len;
  config_node *head;
  int intervals2;
  double ratio;

  path_len=len;
  /* choose two different random points */
  do {
    point1=random() % path_len;
    point2=random() % path_len;
  } while (point1==point2);
  if (point1>point2)
    swap_int(&point1,&point2);
  /* invariant: point1<point2 */
  ratio=
    sqrt(sqr_euclidean_distance(pX[point1],pX[point2]))/(double)DOF;
  intervals2=ratio*intervals;
  head=collision_free(intervals2,pX,point1,point2,NULL);
  if (head!=NULL)
    printf("collision free...\n");
  *ppoint1=point1;
  *ppoint2=point2;
  return(head);
}

/***************************************************************************
  
  smooth1: recursive midpoint (or random) sampling of configuration space

****************************************************************************/

config_node *smooth1(intervals,pX,start,end,head)
     int intervals;
     configuration pX[MAXPATH];
     int start, end;
     config_node *head;
{
  int mid;
  config_node *item, *new_head, *head1, *head2, *local_head;
  int guess, range;
  double ratio;

  if ((end-start) < 2) {
    printf("can't find smoothed path: start=%d end=%d\n", start, end);
    local_head=NULL;
    while (start<=end) {
      CONS(config_node,item,local_head);
      bcopy(pX[start++], item->q, sizeof(configuration));
    }
    return(local_head);
  }

  if ((new_head=collision_free(intervals,pX,start,end,NULL))!=NULL) {
    printf("straight line between: %d and %d is collision-free.\n",
	   start, end);
    return(new_head);
  }
  else {
    printf("straight line between: %d and %d is *NOT* collision-free.\n",
	   start, end);
    /* take random interval - turns out that taking just the mid-point is
     *  good enough, i.e. mid=(start+end)/2
     */
    range=end-start;
    guess=random() % range;
    ratio=(double)guess/(double)range;
    mid=start+guess;
    /* start to mid */
    printf("calling smooth: between start=%d and mid=%d\n", start, mid);
    head1=smooth1((int)(ratio*intervals),pX,start,mid,head);
    /* mid to end */
    printf("calling smooth: between mid+1=%d and end=%d\n", mid+1, end);
    head2=smooth1((int)((1.0-ratio)*intervals),pX,mid+1,end,head1);
    head2=append_config_node(head2,head1);
    return(head2);
  }
}


/***************************************************************************
  function to check if the given segment of path is coliision free 
  and legal configuration or not 
****************************************************************************/

config_node *collision_free(int intervals,
			    configuration pX[MAXPATH],
			    int start, 
			    int end,
			    config_node *head) 
{
  configuration config;
  FPoint kine_config[NBP];
  int i;
  config_node *item;

  for (i=1; i<intervals; i++) {
    INTERPOLATE(pX[start], pX[end], i, intervals, config);

    FK(config,kine_config);

    if(out_ws(kine_config)) {
      free_config_node(head);
      return(NULL);
    }

    if (auto_collision(config, kine_config)) {
      free_config_node(head);
      return(NULL);
    }
    /* collision aintance added */

    if (VC(kine_config, config)) {
      free_config_node(head);
      return(NULL);
    }
    CONS(config_node,item,head);
    bcopy(config,item->q,sizeof(configuration));
  }
  return(head);
}

/***************************************************************************
  Function to free the configuration list
***************************************************************************/

void free_config_node(ptr)
     config_node *ptr;
{
  config_node *item,*tmp;
  
  item=ptr;
  while (item!=NULL) {
    tmp=item;
    item=item->next;
    free(tmp);
  }
}
    
/***************************************************************************
  Function to append a configuration to a list
***************************************************************************/

config_node *append_config_node(head1,head2)
     config_node *head1, *head2;
{
  config_node *item, *new_item;
  
  if (head1==NULL)
    return(head2);
  else {
    item=append_config_node(head1->next, head2);
    new_item=(config_node *)calloc(1,sizeof(config_node));
    bcopy(head1->q,new_item->q,sizeof(configuration));
    new_item->next=item;
    free(head1);
    return(new_item);
  }
}


/***************************************************************************
  Function to compute the length of a list
***************************************************************************/

int list_length(head)
     config_node *head;
{
  config_node *item;
  int i;
  
  item=head;
  i=0;
  while (item!=NULL) {
    i++;
    item=item->next;
  }
  return(i);
}

/***************************************************************************
  Function to swap two integers
***************************************************************************/

void swap_int(p1,p2)
     int *p1, *p2;
{
  static int temp;
  
  temp = *p1;
  *p1 = *p2;
  *p2 = temp;
}

/***************************************************************************
  Function to shift the position of a list of configurations
***************************************************************************/

int shift_path(pX, len, head, point1, point2)
     configuration pX[MAXPATH];
     int len;
     config_node *head;
     int point1, point2;
{
  int i, j, path_len, rest, skip, list_len, k;
  config_node *item;
  
  path_len=len;
  i=list_len=list_length(head);

  /* now shift rest of configuration array over - including end point */
  rest=path_len-point2;
  skip=(point2-point1)-list_len-1;
  if (skip) {
    for (j=0, k=rest-1; j<rest; j++, k--) {
      if (skip>0)
	bcopy(pX[point2+j], pX[point2-skip+j], sizeof(configuration));
      else {
	bcopy(pX[point2+k], pX[point2-skip+k], sizeof(configuration));
      }
    }
  }

  /* copy in reverse list of collision free points to array */
  item=head;
  while (item!=NULL) {
    k=point1+i;
    bcopy(item->q, pX[k], sizeof(configuration));
    i--;
    item=item->next;
  }
  /* set new path dim */
  path_len =(point1+1)+list_len+rest;
  return(path_len);
}

/***************************************************************************
  Compute the square root of euclidean distance
***************************************************************************/
double sqr_euclidean_distance(q1, q2)
     configuration q1, q2;
{
  double sum=0, delta;
  int i;

  for (i=0; i<3; i++) {
    delta=q2[i]-q1[i];
    sum += delta*delta;
  }
  for (i=3; i<DOF; i++) {
    delta=my_ABS(q2[i]-q1[i]);
    if (delta > (0.5*DIM)) {
      delta=DIM-delta;
    }
    sum += delta*delta;
  }

  return(sum);
}


/***************************************************************************
  Smooth the path so that the motion in one time step would be more even
***************************************************************************/

double curvilinear(configuration X[MAXPATH], int path_length) 
{
  register int i;
  double length=0;
  double offset;
  configuration *Xold;
  double *lg;
  int posit;

  printf("Computing curvilinear approximation...\n");
  Xold = (configuration *)calloc(path_length,sizeof(configuration));
  lg = (double *)calloc(path_length,sizeof(double));

  	/* save old path */
  bcopy(X[0], Xold[0], (path_length)*sizeof(configuration));

  	/* compute partial lengths of old path */
  for (i=0; i<path_length-1; i++) {
    lg[i] = length;
    length += sqrt(sqr_euclidean_distance(Xold[i],Xold[i+1]));
  }
  lg[path_length-1] = length;

  posit = 0;
  for (i=1; i<path_length-1; i++) {
    while ((offset = (length*i/(double)(path_length-1)-lg[posit+1])) > 0)
      posit++;
/*
    
    for (j=0; j<3; j++) {
      X[i][j] = GRID(Xold[posit+1][j] +
	offset/(lg[posit+1]-lg[posit]) * (Xold[posit+1][j] - Xold[posit][j]));

    }

    for (j=3; j<DOF; j++) {
      difference = (Xold[posit+1][j] - Xold[posit][j]);      
      if (my_ABS(difference) > (0.5*DIM)) {
	if (difference > 0.0) 
	  difference = DIM  - difference;
	else
	  difference = DIM  + difference;
      }
      X[i][j] = GRID(Xold[posit+1][j] +
	offset/(lg[posit+1]-lg[posit]) * difference);
      if (X[i][j] >= DIM) 
	X[i][j] -= DIM;
      if (X[i][j] < 0)
	X[i][j] += DIM;
    }
*/
  }
  free(Xold);
  free(lg);
  return(length);
}


/***************************************************************************
  Smooth the path so that the motion in one time step would be more even
***************************************************************************/

int CurvePath(configuration X[MAXPATH], int dim) {
  int i,j,k;
  double length=0.0;
  double offset,difference;
  configuration *Xold;
  double *lg;
  int posit;
  FPoint kine_conf[NBP];

  if (dim<3) return 1;
  Xold = (configuration *)calloc(dim,sizeof(configuration));
  lg = (double *)calloc(dim,sizeof(double));

  bcopy(X[0], Xold[0],dim*sizeof(configuration));
  for (i=0; i<dim-1; i++) {
    lg[i] = length;
    length += sqrt(sqr_euclidean_distance(Xold[i],Xold[i+1]));
  }
  lg[dim-1] = length;
  
  posit = 0;
  for (i=0; i<dim-1; i++) {
    while ((offset = (length*i/(double)(dim-1)-lg[posit+1])) > 0)
      posit++;
    
    for(j=0; j<3; j++) {
     X[i][j] = Xold[posit+1][j] +
      (int)(offset/(lg[posit+1]-lg[posit])*(Xold[posit+1][j]-Xold[posit][j]));
    }

    for (j=3; j<DOF; j++) {
	difference = (Xold[posit+1][j] - Xold[posit][j]);      
	if (my_ABS(difference) > (0.5*DIM)) {
	  if (difference > 0.0) 
	    difference = DIM  - difference;
	  else
	    difference = DIM  + difference;
	}
	
	X[i][j] = Xold[posit+1][j] +
	  (int)(offset/(lg[posit+1]-lg[posit]) * difference);
	if (X[i][j] >= DIM) 
	  X[i][j] -= DIM;
	if (X[i][j] < 0)
	  X[i][j] += DIM;
      } 

    FK(X[i],kine_conf);
    if(out_ws(kine_conf) ||  auto_collision(X[i],kine_conf)
		     || VC(kine_conf, X[i])) {
      bcopy(Xold[0],X[0],dim*sizeof(configuration));
      k=(int)(dim/2);
      free(Xold);
      free(lg);
      CurvePath((configuration *)&(X[0][0]),k);
      CurvePath((configuration *)&(X[k][0]),dim-k);
      return(0);
    }
  }
  free(Xold);
  free(lg);
  return(1);
}

